@@ -1,8 +1,8 @@
-#include "pysam.h"
+#include "bcftools.pysam.h"
 
 /*  vcfroh.c -- HMM model for detecting runs of autozygosity.
 
-    Copyright (C) 2013-2015 Genome Research Ltd.
+    Copyright (C) 2013-2018 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -28,6 +28,7 @@ THE SOFTWARE.  */
 #include <unistd.h>
 #include <getopt.h>
 #include <math.h>
+#include <inttypes.h>
 #include <htslib/vcf.h>
 #include <htslib/synced_bcf_reader.h>
 #include <htslib/kstring.h>
@@ -37,6 +38,7 @@ THE SOFTWARE.  */
 #include "bcftools.h"
 #include "HMM.h"
 #include "smpl_ilist.h"
+#include "filter.h"
 
 #define STATE_HW 0        // normal state, follows Hardy-Weinberg allele frequencies
 #define STATE_AZ 1        // autozygous state
@@ -45,6 +47,11 @@ THE SOFTWARE.  */
 #define OUTPUT_RG (1<<2)
 #define OUTPUT_GZ (1<<3)
 
+// Logic of the filters: include or exclude sites which match the filters?
+#define FLT_INCLUDE 1
+#define FLT_EXCLUDE 2
+
+
 /** Genetic map */
 typedef struct
 {
@@ -96,6 +103,9 @@ typedef struct _args_t
     int32_t skip_rid, prev_rid, prev_pos;
 
     int ntot;                   // some stats to detect if things didn't go wrong
+    int nno_af;                 // number of sites rejected because AF could not be determined
+    int nfiltered;              // .. because of filters
+    int nnot_biallelic, ndup;
     smpl_t *smpl;               // HMM data for each sample
     smpl_ilist_t *af_smpl;      // list of samples to estimate AF from (--estimate-AF)
     smpl_ilist_t *roh_smpl;     // list of samples to analyze (--samples, --samples-file)
@@ -105,6 +115,10 @@ typedef struct _args_t
     int argc, fake_PLs, snps_only, vi_training, samples_is_file, output_type, skip_homref, n_threads;
     BGZF *out;
     kstring_t str;
+
+    int filter_logic;
+    filter_t *filter;
+    char *filter_str;
 }
 args_t;
 
@@ -118,6 +132,11 @@ void *smalloc(size_t size)
     return mem;
 }
 
+static inline int max255(int i)
+{
+    return i < 256 ? i : 255;
+}
+
 static void init_data(args_t *args)
 {
     int i;
@@ -127,6 +146,9 @@ static void init_data(args_t *args)
 
     if ( !bcf_hdr_nsamples(args->hdr) ) error("No samples in the VCF?\n");
 
+    if ( args->filter_str )
+        args->filter = filter_init(args->hdr, args->filter_str);
+
     if ( !args->fake_PLs )
     {
         args->pl_hdr_id = bcf_hdr_id2int(args->hdr, BCF_DT_ID, "PL");
@@ -141,7 +163,7 @@ static void init_data(args_t *args)
         if ( !strncmp("GT,",args->estimate_AF,3) ) args->estimate_AF += 3;
         else if ( !strncmp("PL,",args->estimate_AF,3) ) { args->estimate_AF += 3; args->af_from_PL = 1; }
         if ( strcmp("-",args->estimate_AF) )
-            args->af_smpl = smpl_ilist_init(args->hdr, args->estimate_AF, 1, SMPL_NONE);
+            args->af_smpl = smpl_ilist_init(args->hdr, args->estimate_AF, 1, SMPL_NONE|SMPL_VERBOSE);
     }
 
     if ( args->estimate_AF || args->fake_PLs )
@@ -166,7 +188,7 @@ static void init_data(args_t *args)
             error("Error: The FORMAT/GT tag not found in the header\n");
     }
 
-    args->roh_smpl = smpl_ilist_init(args->hdr, args->samples, args->samples_is_file, SMPL_NONE);
+    args->roh_smpl = smpl_ilist_init(args->hdr, args->samples, args->samples_is_file, SMPL_NONE|SMPL_VERBOSE);
     if ( args->samples )
     {
         // we may be able to subset to a few samples, for a text VCF this can be a major speedup
@@ -243,11 +265,11 @@ static void init_data(args_t *args)
         if ( args->nbuf_olap<0 )
             args->nbuf_olap = args->nbuf_max*0.01;
     }
-    fprintf(pysam_stderr,"Number of target samples: %d\n", args->roh_smpl->n);
-    fprintf(pysam_stderr,"Number of --estimate-AF samples: %d\n", args->af_smpl ? args->af_smpl->n : (args->estimate_AF ? bcf_hdr_nsamples(args->hdr) : 0));
-    fprintf(pysam_stderr,"Number of sites in the buffer/overlap: ");
-    if ( args->nbuf_max ) fprintf(pysam_stderr,"%d/%d\n", args->nbuf_max,args->nbuf_olap);
-    else fprintf(pysam_stderr,"unlimited\n");
+    fprintf(bcftools_stderr,"Number of target samples: %d\n", args->roh_smpl->n);
+    fprintf(bcftools_stderr,"Number of --estimate-AF samples: %d\n", args->af_smpl ? args->af_smpl->n : (args->estimate_AF ? bcf_hdr_nsamples(args->hdr) : 0));
+    fprintf(bcftools_stderr,"Number of sites in the buffer/overlap: ");
+    if ( args->nbuf_max ) fprintf(bcftools_stderr,"%d/%d\n", args->nbuf_max,args->nbuf_olap);
+    else fprintf(bcftools_stderr,"unlimited\n");
 
     args->smpl = (smpl_t*) calloc(args->roh_smpl->n,sizeof(smpl_t));
 
@@ -266,7 +288,7 @@ static void init_data(args_t *args)
     else if ( args->rec_rate > 0 )
         hmm_set_tprob_func(args->hmm, set_tprob_rrate, args);
 
-    args->out = bgzf_open(strcmp("pysam_stdout",args->output_fname)?args->output_fname:"-", args->output_type&OUTPUT_GZ ? "wg" : "wu"); 
+    args->out = bgzf_open(strcmp("bcftools_stdout",args->output_fname)?args->output_fname:"-", args->output_type&OUTPUT_GZ ? "wg" : "wu"); 
     if ( !args->out ) error("Failed to open %s: %s\n", args->output_fname, strerror(errno));
 
     // print header
@@ -320,6 +342,7 @@ static void init_data(args_t *args)
 
 static void destroy_data(args_t *args)
 {
+    if ( args->filter ) filter_destroy(args->filter);
     if ( bgzf_close(args->out)!=0 ) error("Error: close failed .. %s\n", args->output_fname);
     int i;
     for (i=0; i<args->roh_smpl->n; i++)
@@ -369,7 +392,7 @@ static int load_genmap(args_t *args, const char *chr)
 
     hts_getline(fp, KS_SEP_LINE, &str);
     if ( strcmp(str.s,"position COMBINED_rate(cM/Mb) Genetic_Map(cM)") )
-        error("Unexpected header, found:\n\t[%s], but expected:\n\t[position COMBINED_rate(cM/Mb) Genetic_Map(cM)]\n", fname, str.s);
+        error("Unexpected header in %s, found:\n\t[%s], but expected:\n\t[position COMBINED_rate(cM/Mb) Genetic_Map(cM)]\n", fname, str.s);
 
     args->ngenmap = args->igenmap = 0;
     while ( hts_getline(fp, KS_SEP_LINE, &str) > 0 )
@@ -528,6 +551,8 @@ static void flush_viterbi(args_t *args, int ismpl)
                         smpl->rg.state = 1;
                         smpl->rg.beg = smpl->sites[i];
                         smpl->rg.rid = args->prev_rid;
+                        smpl->rg.qual  = qual;
+                        smpl->rg.nqual = 1;
                     }
                 }
                 else if ( state )
@@ -731,9 +756,9 @@ int estimate_AF_from_PL(args_t *args, bcf_fmt_t *fmt_pl, int ial, double *alt_fr
                 if ( p[irr]<0 || p[ira]<0 || p[iaa]<0 ) continue;    /* missing value */ \
                 if ( p[irr]==p[ira] && p[irr]==p[iaa] ) continue;    /* all values are the same */ \
                 double prob[3], norm = 0; \
-                prob[0] = p[irr] < (type_t)256 ? args->pl2p[ p[irr] ] : args->pl2p[255]; \
-                prob[1] = p[ira] < (type_t)256 ? args->pl2p[ p[ira] ] : args->pl2p[255]; \
-                prob[2] = p[iaa] < (type_t)256 ? args->pl2p[ p[iaa] ] : args->pl2p[255]; \
+                prob[0] = args->pl2p[ max255(p[irr]) ]; \
+                prob[1] = args->pl2p[ max255(p[ira]) ]; \
+                prob[2] = args->pl2p[ max255(p[iaa]) ]; \
                 for (j=0; j<3; j++) norm += prob[j]; \
                 for (j=0; j<3; j++) prob[j] /= norm; \
                 af += 0.5*prob[1] + prob[2]; \
@@ -744,7 +769,7 @@ int estimate_AF_from_PL(args_t *args, bcf_fmt_t *fmt_pl, int ial, double *alt_fr
             case BCF_BT_INT8:  BRANCH(int8_t); break;
             case BCF_BT_INT16: BRANCH(int16_t); break;
             case BCF_BT_INT32: BRANCH(int32_t); break;
-            default: fprintf(pysam_stderr,"Unknown format type for PL: %s:%d .. fmt->type=%d\n", __FILE__,__LINE__, fmt_pl->type); exit(1);
+            default: fprintf(bcftools_stderr,"Unknown format type for PL: %s:%d .. fmt->type=%d\n", __FILE__,__LINE__, fmt_pl->type); exit(1);
         }
         #undef BRANCH
     }
@@ -761,9 +786,9 @@ int estimate_AF_from_PL(args_t *args, bcf_fmt_t *fmt_pl, int ial, double *alt_fr
                 if ( p[irr]<0 || p[ira]<0 || p[iaa]<0 ) continue;    /* missing value */ \
                 if ( p[irr]==p[ira] && p[irr]==p[iaa] ) continue;    /* all values are the same */ \
                 double prob[3], norm = 0; \
-                prob[0] = p[irr] < (type_t)256 ? args->pl2p[ p[irr] ] : args->pl2p[255]; \
-                prob[1] = p[ira] < (type_t)256 ? args->pl2p[ p[ira] ] : args->pl2p[255]; \
-                prob[2] = p[iaa] < (type_t)256 ? args->pl2p[ p[iaa] ] : args->pl2p[255]; \
+                prob[0] = args->pl2p[ max255(p[irr]) ]; \
+                prob[1] = args->pl2p[ max255(p[ira]) ]; \
+                prob[2] = args->pl2p[ max255(p[iaa]) ]; \
                 for (j=0; j<3; j++) norm += prob[j]; \
                 for (j=0; j<3; j++) prob[j] /= norm; \
                 af += 0.5*prob[1] + prob[2]; \
@@ -774,7 +799,7 @@ int estimate_AF_from_PL(args_t *args, bcf_fmt_t *fmt_pl, int ial, double *alt_fr
             case BCF_BT_INT8:  BRANCH(int8_t); break;
             case BCF_BT_INT16: BRANCH(int16_t); break;
             case BCF_BT_INT32: BRANCH(int32_t); break;
-            default: fprintf(pysam_stderr,"Unknown format type for PL: %s:%d .. fmt->type=%d\n", __FILE__,__LINE__, fmt_pl->type); exit(1);
+            default: fprintf(bcftools_stderr,"Unknown format type for PL: %s:%d .. fmt->type=%d\n", __FILE__,__LINE__, fmt_pl->type); exit(1);
         }
         #undef BRANCH
     }
@@ -809,7 +834,7 @@ int process_line(args_t *args, bcf1_t *line, int ial)
         if ( ret>0 )
             alt_freq = args->AFs[ial-1];
         if ( ret==-2 )
-            error("Type mismatch for INFO/%s tag at %s:%d\n", args->af_tag, bcf_seqname(args->hdr,line), line->pos+1);
+            error("Type mismatch for INFO/%s tag at %s:%"PRId64"\n", args->af_tag, bcf_seqname(args->hdr,line), (int64_t) line->pos+1);
     }
     else if ( args->af_fname ) 
     {
@@ -854,8 +879,9 @@ int process_line(args_t *args, bcf1_t *line, int ial)
             alt_freq = (double) AC/AN;
     }
 
-    if ( ret<0 ) return ret;
-    if ( alt_freq==0.0 ) return -1;
+    if ( args->dflt_AF>0 && (ret<0 || alt_freq==0.0) ) alt_freq = args->dflt_AF;
+    else if ( ret<0 ) { args->nno_af++; return ret; }
+    else if ( alt_freq==0.0 ) { args->nno_af++; return -1; }
 
     int irr = bcf_alleles2gt(0,0), ira = bcf_alleles2gt(0,ial), iaa = bcf_alleles2gt(ial,ial);
     if ( args->fake_PLs )
@@ -907,15 +933,15 @@ int process_line(args_t *args, bcf1_t *line, int ial)
                 type_t *p = (type_t*)fmt_pl->p + fmt_pl->n*ismpl; \
                 if ( p[irr]<0 || p[ira]<0 || p[iaa]<0 ) continue;    /* missing value */ \
                 if ( p[irr]==p[ira] && p[irr]==p[iaa] ) continue;    /* all values are the same */ \
-                pdg[0] = p[irr] < (type_t)256 ? args->pl2p[ p[irr] ] : args->pl2p[255]; \
-                pdg[1] = p[ira] < (type_t)256 ? args->pl2p[ p[ira] ] : args->pl2p[255]; \
-                pdg[2] = p[iaa] < (type_t)256 ? args->pl2p[ p[iaa] ] : args->pl2p[255]; \
+                pdg[0] = args->pl2p[ max255(p[irr]) ]; \
+                pdg[1] = args->pl2p[ max255(p[ira]) ]; \
+                pdg[2] = args->pl2p[ max255(p[iaa]) ]; \
             }
             switch (fmt_pl->type) {
                 case BCF_BT_INT8:  BRANCH(int8_t); break;
                 case BCF_BT_INT16: BRANCH(int16_t); break;
                 case BCF_BT_INT32: BRANCH(int32_t); break;
-                default: fprintf(pysam_stderr,"Unknown format type for PL: %s:%d .. fmt->type=%d\n", __FILE__,__LINE__, fmt_pl->type); exit(1);
+                default: fprintf(bcftools_stderr,"Unknown format type for PL: %s:%d .. fmt->type=%d\n", __FILE__,__LINE__, fmt_pl->type); exit(1);
             }
             #undef BRANCH
         }
@@ -932,7 +958,7 @@ int process_line(args_t *args, bcf1_t *line, int ial)
         {
             hts_expand(uint32_t,smpl->nsites+1,smpl->msites,smpl->sites);
             smpl->eprob = (double*) realloc(smpl->eprob,sizeof(*smpl->eprob)*smpl->msites*2);
-            if ( !smpl->eprob ) error("Error: failed to alloc %d bytes\n", sizeof(*smpl->eprob)*smpl->msites*2);
+            if ( !smpl->eprob ) error("Error: failed to alloc %"PRIu64" bytes\n", (uint64_t)(sizeof(*smpl->eprob)*smpl->msites*2));
         }
         
         // Calculate emission probabilities P(D|AZ) and P(D|HW)
@@ -970,12 +996,11 @@ static void vcfroh(args_t *args, bcf1_t *line)
         for (i=0; i<args->roh_smpl->n; i++) flush_viterbi(args, i);
         return; 
     }
-    args->ntot++;
 
     // Skip unwanted lines, for simplicity we consider only biallelic sites 
     if ( line->rid == args->skip_rid ) return;
-    if ( line->n_allele==1 ) return;    // no ALT allele
-    if ( line->n_allele > 3 ) return;   // cannot be bi-allelic, even with <*>
+    if ( line->n_allele==1 ) { args->nnot_biallelic++; return; }   // no ALT allele
+    if ( line->n_allele > 3 ) { args->nnot_biallelic++; return; }   // cannot be bi-allelic, even with <*>
 
     // This can be raw callable VCF with the symbolic unseen allele <*>
     int ial = 0;
@@ -983,7 +1008,7 @@ static void vcfroh(args_t *args, bcf1_t *line)
         if ( !strcmp("<*>",line->d.allele[i]) ) { ial = i; break; }
     if ( ial==0 )    // normal VCF, the symbolic allele is not present
     {
-        if ( line->n_allele!=2 ) return;    // not biallelic
+        if ( line->n_allele!=2 ) { args->nnot_biallelic++; return; }   // not biallelic
         ial = 1;
     }
     else
@@ -1017,11 +1042,15 @@ static void vcfroh(args_t *args, bcf1_t *line)
         args->prev_pos = line->pos;
         skip_rid = load_genmap(args, bcf_seqname(args->hdr,line));
     }
-    else if ( args->prev_pos == line->pos ) return;     // skip duplicate positions
+    else if ( args->prev_pos == line->pos ) 
+    {
+        args->ndup++;
+        return;     // skip duplicate positions
+    }
 
     if ( skip_rid )
     {
-        fprintf(pysam_stderr,"Skipping the sequence, no genmap for %s\n", bcf_seqname(args->hdr,line));
+        fprintf(bcftools_stderr,"Skipping the sequence, no genmap for %s\n", bcf_seqname(args->hdr,line));
         args->skip_rid = line->rid;
         return;
     }
@@ -1037,41 +1066,43 @@ static void vcfroh(args_t *args, bcf1_t *line)
 
 static void usage(args_t *args)
 {
-    fprintf(pysam_stderr, "\n");
-    fprintf(pysam_stderr, "About:   HMM model for detecting runs of autozygosity.\n");
-    fprintf(pysam_stderr, "Usage:   bcftools roh [options] <in.vcf.gz>\n");
-    fprintf(pysam_stderr, "\n");
-    fprintf(pysam_stderr, "General Options:\n");
-    fprintf(pysam_stderr, "        --AF-dflt <float>              if AF is not known, use this allele frequency [skip]\n");
-    fprintf(pysam_stderr, "        --AF-tag <TAG>                 use TAG for allele frequency\n");
-    fprintf(pysam_stderr, "        --AF-file <file>               read allele frequencies from file (CHR\\tPOS\\tREF,ALT\\tAF)\n");
-    fprintf(pysam_stderr, "    -b  --buffer-size <int[,int]>      buffer size and the number of overlapping sites, 0 for unlimited [0]\n");
-    fprintf(pysam_stderr, "                                           If the first number is negative, it is interpreted as the maximum memory to\n");
-    fprintf(pysam_stderr, "                                           use, in MB. The default overlap is set to roughly 1%% of the buffer size.\n");
-    fprintf(pysam_stderr, "    -e, --estimate-AF [TAG],<file>     estimate AF from FORMAT/TAG (GT or PL) of all samples (\"-\") or samples listed\n");
-    fprintf(pysam_stderr, "                                            in <file>. If TAG is not given, the frequency is estimated from GT by default\n");
-    fprintf(pysam_stderr, "    -G, --GTs-only <float>             use GTs and ignore PLs, instead using <float> for PL of the two least likely genotypes.\n");
-    fprintf(pysam_stderr, "                                           Safe value to use is 30 to account for GT errors.\n");
-    fprintf(pysam_stderr, "    -i, --ignore-homref                skip hom-ref genotypes (0/0)\n");
-    fprintf(pysam_stderr, "    -I, --skip-indels                  skip indels as their genotypes are enriched for errors\n");
-    fprintf(pysam_stderr, "    -m, --genetic-map <file>           genetic map in IMPUTE2 format, single file or mask, where string \"{CHROM}\"\n");
-    fprintf(pysam_stderr, "                                           is replaced with chromosome name\n");
-    fprintf(pysam_stderr, "    -M, --rec-rate <float>             constant recombination rate per bp\n");
-    fprintf(pysam_stderr, "    -o, --output <file>                write output to a file [standard output]\n");
-    fprintf(pysam_stderr, "    -O, --output-type [srz]            output s:per-site, r:regions, z:compressed [sr]\n");
-    fprintf(pysam_stderr, "    -r, --regions <region>             restrict to comma-separated list of regions\n");
-    fprintf(pysam_stderr, "    -R, --regions-file <file>          restrict to regions listed in a file\n");
-    fprintf(pysam_stderr, "    -s, --samples <list>               list of samples to analyze [all samples]\n");
-    fprintf(pysam_stderr, "    -S, --samples-file <file>          file of samples to analyze [all samples]\n");
-    fprintf(pysam_stderr, "    -t, --targets <region>             similar to -r but streams rather than index-jumps\n");
-    fprintf(pysam_stderr, "    -T, --targets-file <file>          similar to -R but streams rather than index-jumps\n");
-    fprintf(pysam_stderr, "        --threads <int>                number of extra decompression threads [0]\n");
-    fprintf(pysam_stderr, "\n");
-    fprintf(pysam_stderr, "HMM Options:\n");
-    fprintf(pysam_stderr, "    -a, --hw-to-az <float>             P(AZ|HW) transition probability from HW (Hardy-Weinberg) to AZ (autozygous) state [6.7e-8]\n");
-    fprintf(pysam_stderr, "    -H, --az-to-hw <float>             P(HW|AZ) transition probability from AZ to HW state [5e-9]\n");
-    fprintf(pysam_stderr, "    -V, --viterbi-training <float>     estimate HMM parameters, <float> is the convergence threshold, e.g. 1e-10 (experimental)\n");
-    fprintf(pysam_stderr, "\n");
+    fprintf(bcftools_stderr, "\n");
+    fprintf(bcftools_stderr, "About:   HMM model for detecting runs of autozygosity.\n");
+    fprintf(bcftools_stderr, "Usage:   bcftools roh [options] <in.vcf.gz>\n");
+    fprintf(bcftools_stderr, "\n");
+    fprintf(bcftools_stderr, "General Options:\n");
+    fprintf(bcftools_stderr, "        --AF-dflt <float>              if AF is not known, use this allele frequency [skip]\n");
+    fprintf(bcftools_stderr, "        --AF-tag <TAG>                 use TAG for allele frequency\n");
+    fprintf(bcftools_stderr, "        --AF-file <file>               read allele frequencies from file (CHR\\tPOS\\tREF\\tALT\\tAF)\n");
+    fprintf(bcftools_stderr, "    -b  --buffer-size <int[,int]>      buffer size and the number of overlapping sites, 0 for unlimited [0]\n");
+    fprintf(bcftools_stderr, "                                           If the first number is negative, it is interpreted as the maximum memory to\n");
+    fprintf(bcftools_stderr, "                                           use, in MB. The default overlap is set to roughly 1%% of the buffer size.\n");
+    fprintf(bcftools_stderr, "    -e, --estimate-AF [TAG],<file>     estimate AF from FORMAT/TAG (GT or PL) of all samples (\"-\") or samples listed\n");
+    fprintf(bcftools_stderr, "                                            in <file>. If TAG is not given, the frequency is estimated from GT by default\n");
+    fprintf(bcftools_stderr, "        --exclude <expr>               exclude sites for which the expression is true\n");
+    fprintf(bcftools_stderr, "    -G, --GTs-only <float>             use GTs and ignore PLs, instead using <float> for PL of the two least likely genotypes.\n");
+    fprintf(bcftools_stderr, "                                           Safe value to use is 30 to account for GT errors.\n");
+    fprintf(bcftools_stderr, "        --include <expr>               select sites for which the expression is true\n");
+    fprintf(bcftools_stderr, "    -i, --ignore-homref                skip hom-ref genotypes (0/0)\n");
+    fprintf(bcftools_stderr, "    -I, --skip-indels                  skip indels as their genotypes are enriched for errors\n");
+    fprintf(bcftools_stderr, "    -m, --genetic-map <file>           genetic map in IMPUTE2 format, single file or mask, where string \"{CHROM}\"\n");
+    fprintf(bcftools_stderr, "                                           is replaced with chromosome name\n");
+    fprintf(bcftools_stderr, "    -M, --rec-rate <float>             constant recombination rate per bp\n");
+    fprintf(bcftools_stderr, "    -o, --output <file>                write output to a file [standard output]\n");
+    fprintf(bcftools_stderr, "    -O, --output-type [srz]            output s:per-site, r:regions, z:compressed [sr]\n");
+    fprintf(bcftools_stderr, "    -r, --regions <region>             restrict to comma-separated list of regions\n");
+    fprintf(bcftools_stderr, "    -R, --regions-file <file>          restrict to regions listed in a file\n");
+    fprintf(bcftools_stderr, "    -s, --samples <list>               list of samples to analyze [all samples]\n");
+    fprintf(bcftools_stderr, "    -S, --samples-file <file>          file of samples to analyze [all samples]\n");
+    fprintf(bcftools_stderr, "    -t, --targets <region>             similar to -r but streams rather than index-jumps\n");
+    fprintf(bcftools_stderr, "    -T, --targets-file <file>          similar to -R but streams rather than index-jumps\n");
+    fprintf(bcftools_stderr, "        --threads <int>                use multithreading with <int> worker threads [0]\n");
+    fprintf(bcftools_stderr, "\n");
+    fprintf(bcftools_stderr, "HMM Options:\n");
+    fprintf(bcftools_stderr, "    -a, --hw-to-az <float>             P(AZ|HW) transition probability from HW (Hardy-Weinberg) to AZ (autozygous) state [6.7e-8]\n");
+    fprintf(bcftools_stderr, "    -H, --az-to-hw <float>             P(HW|AZ) transition probability from AZ to HW state [5e-9]\n");
+    fprintf(bcftools_stderr, "    -V, --viterbi-training <float>     estimate HMM parameters, <float> is the convergence threshold, e.g. 1e-10 (experimental)\n");
+    fprintf(bcftools_stderr, "\n");
     exit(1);
 }
 
@@ -1091,6 +1122,8 @@ int main_vcfroh(int argc, char *argv[])
         {"AF-tag",1,0,0},
         {"AF-file",1,0,1},
         {"AF-dflt",1,0,2},
+        {"include",1,0,3},
+        {"exclude",1,0,4},
         {"buffer-size",1,0,'b'},
         {"ignore-homref",0,0,'i'},
         {"estimate-AF",1,0,'e'},
@@ -1123,6 +1156,8 @@ int main_vcfroh(int argc, char *argv[])
                 args->dflt_AF = strtod(optarg,&tmp);
                 if ( *tmp ) error("Could not parse: --AF-dflt %s\n", optarg);
                 break;
+            case 3: args->filter_str = optarg; args->filter_logic = FLT_INCLUDE; break;
+            case 4: args->filter_str = optarg; args->filter_logic = FLT_EXCLUDE; break;
             case 'o': args->output_fname = optarg; break;
             case 'O': 
                 if ( strchr(optarg,'s') || strchr(optarg,'S') ) args->output_type |= OUTPUT_ST;
@@ -1169,8 +1204,8 @@ int main_vcfroh(int argc, char *argv[])
             default: error("Unknown argument: %s\n", optarg);
         }
     }
-    if ( !args->output_fname ) args->output_fname = "pysam_stdout";
-    if ( !args->output_type ) args->output_type = OUTPUT_ST|OUTPUT_RG;
+    if ( !args->output_fname ) args->output_fname = "bcftools_stdout";
+    if ( !args->output_type || args->output_type==OUTPUT_GZ ) args->output_type |= OUTPUT_ST|OUTPUT_RG;
     char *fname = NULL;
     if ( optind==argc )
     {
@@ -1180,8 +1215,8 @@ int main_vcfroh(int argc, char *argv[])
     else fname = argv[optind];
 
     if ( args->vi_training && args->buffer_size ) error("Error: cannot use -b with -V\n");
-    if ( args->t2AZ<0 || args->t2AZ>1 ) error("Error: The parameter --hw-to-az is not in [0,1]\n", args->t2AZ);
-    if ( args->t2HW<0 || args->t2HW>1 ) error("Error: The parameter --az-to-hw is not in [0,1]\n", args->t2HW);
+    if ( args->t2AZ<0 || args->t2AZ>1 ) error("Error: The parameter --hw-to-az is not in [0,1] .. %e\n", args->t2AZ);
+    if ( args->t2HW<0 || args->t2HW>1 ) error("Error: The parameter --az-to-hw is not in [0,1] .. %e\n", args->t2HW);
     if ( naf_opts>1 ) error("Error: The options --AF-tag, --AF-file and -e are mutually exclusive\n");
     if ( args->af_fname && args->targets_list ) error("Error: The options --AF-file and -t are mutually exclusive\n");
     if ( args->regions_list )
@@ -1201,22 +1236,34 @@ int main_vcfroh(int argc, char *argv[])
     }
     if ( args->n_threads && bcf_sr_set_threads(args->files, args->n_threads)<0)
         error("Failed to create threads\n");
-    if ( !bcf_sr_add_reader(args->files, fname) ) error("Failed to open %s: %s\n", fname,bcf_sr_strerror(args->files->errnum));
+    if ( !bcf_sr_add_reader(args->files, fname) ) error("Failed to read from %s: %s\n", !strcmp("-",fname)?"standard input":fname,bcf_sr_strerror(args->files->errnum));
 
     init_data(args);
     while ( bcf_sr_next_line(args->files) )
     {
-        vcfroh(args, args->files->readers[0].buffer[0]);
+        args->ntot++;
+        bcf1_t *line = bcf_sr_get_line(args->files,0);
+        if ( args->filter )
+        {
+            int pass = filter_test(args->filter, line, NULL);
+            if ( args->filter_logic & FLT_EXCLUDE ) pass = pass ? 0 : 1;
+            if ( !pass ) { args->nfiltered++; continue; }
+        }
+        vcfroh(args, line);
     }
     vcfroh(args, NULL);
     int i, nmin = 0;
     for (i=0; i<args->roh_smpl->n; i++)
         if ( !i || args->smpl[i].nused < nmin ) nmin = args->smpl[i].nused;
-    fprintf(pysam_stderr,"Number of lines total/processed: %d/%d\n", args->ntot,nmin);
+    if ( args->af_fname )
+        fprintf(bcftools_stderr,"Number of lines overlapping with --AF-file/processed: %d/%d\n", args->ntot,nmin);
+    else
+        fprintf(bcftools_stderr,"Number of lines total/processed: %d/%d\n", args->ntot,nmin);
+    fprintf(bcftools_stderr,"Number of lines filtered/no AF/not biallelic/dup: %d/%d/%d/%d\n", args->nfiltered,args->nno_af,args->nnot_biallelic,args->ndup);
     if ( nmin==0 )
     {
-        fprintf(pysam_stderr,"No usable sites were found.");
-        if ( !naf_opts && !args->dflt_AF ) fprintf(pysam_stderr, " Consider using one of the AF options.\n");
+        fprintf(bcftools_stderr,"No usable sites were found.\n");
+        if ( !naf_opts && !args->dflt_AF ) fprintf(bcftools_stderr, " Consider using one of the AF options.\n");
     }
     destroy_data(args);
     free(args);
